package com.marvinelsen.chinese.transliteration

import io.kotest.core.spec.style.ShouldSpec
import io.kotest.datatest.withData
import io.kotest.matchers.shouldBe

class SyllableTest : ShouldSpec({
    context("from pinyin with tone numbers") {
        withData(
            nameFn = { "'${it.first}' -> ${it.second}" },
            "sheng1" to Syllable("sheng", Tone.FIRST),
            "zhi2" to Syllable("zhi", Tone.SECOND),
            "ka3" to Syllable("ka", Tone.THIRD),
            "yao4" to Syllable("yao", Tone.FORTH),
            "me5" to Syllable("me", Tone.FIFTH),
            "Me5" to Syllable("Me", Tone.FIFTH),
            "nv3" to Syllable("nv", Tone.THIRD),
            "nü3" to Syllable("nü", Tone.THIRD),
            "nu:3" to Syllable("nu:", Tone.THIRD),
        ) { (pinyinWithNumber, expectedSyllable) ->
            Syllable.fromPinyinWithToneNumber(pinyinWithNumber) shouldBe expectedSyllable
        }
    }

    context("from zhuyin") {
        withData(
            nameFn = { "'${it.first}' -> ${it.second}" },
            "ㄕㄥ" to Syllable("sheng", Tone.FIRST),
            "ㄓˊ" to Syllable("zhi", Tone.SECOND),
            "ㄎㄚˇ" to Syllable("ka", Tone.THIRD),
            "ㄧㄠˋ" to Syllable("yao", Tone.FORTH),
            "ㄇㄜ˙" to Syllable("me", Tone.FIFTH),
            "˙ㄇㄜ" to Syllable("me", Tone.FIFTH),
            "ㄋㄩˇ" to Syllable("nü", Tone.THIRD),
        ) { (zhuyin, expectedSyllable) ->
            Syllable.fromZhuyin(zhuyin) shouldBe expectedSyllable
        }
    }

    context("from invalid pinyin with tone numbers") {
    }

    context("from invalid zhuyin") {
    }

    context("format to zhuyin") {
        withData(
            nameFn = { "${it.first.format(TransliterationSystem.PINYIN_WITH_TONE_NUMBERS)} -> ${it.second}" },
            Syllable("sheng", Tone.FIRST) to "ㄕㄥ",
            Syllable("zhi", Tone.SECOND) to "ㄓˊ",
            Syllable("ka", Tone.THIRD) to "ㄎㄚˇ",
            Syllable("yao", Tone.FORTH) to "ㄧㄠˋ",
            Syllable("me", Tone.FIFTH) to "˙ㄇㄜ",
            Syllable("nü", Tone.THIRD) to "ㄋㄩˇ",
            Syllable("nu:", Tone.THIRD) to "ㄋㄩˇ",
            Syllable("nv", Tone.THIRD) to "ㄋㄩˇ",
        ) { (syllable, expectedZhuyin) ->
            syllable.format(TransliterationSystem.ZHUYIN) shouldBe expectedZhuyin
        }
    }

    context("format to pinyin with tone numbers") {
        withData(
            nameFn = { "${it.first.format(TransliterationSystem.PINYIN_WITH_TONE_NUMBERS)} -> ${it.second}" },
            Syllable("sheng", Tone.FIRST) to "sheng1",
            Syllable("zhi", Tone.SECOND) to "zhi2",
            Syllable("ka", Tone.THIRD) to "ka3",
            Syllable("yao", Tone.FORTH) to "yao4",
            Syllable("me", Tone.FIFTH) to "me5",
            Syllable("nü", Tone.THIRD) to "nü3",
            Syllable("nu:", Tone.THIRD) to "nu:3",
            Syllable("nv", Tone.THIRD) to "nv3",
        ) { (syllable, expectedPinyinWithToneNumbers) ->
            syllable.format(TransliterationSystem.PINYIN_WITH_TONE_NUMBERS) shouldBe expectedPinyinWithToneNumbers
        }
    }

    context("format to pinyin with tone marks") {
        withData(
            nameFn = { "${it.first.format(TransliterationSystem.PINYIN_WITH_TONE_NUMBERS)} -> ${it.second}" },
            Syllable("sheng", Tone.FIRST) to "shēng",
            Syllable("zhi", Tone.SECOND) to "zhí",
            Syllable("ka", Tone.THIRD) to "kǎ",
            Syllable("yao", Tone.FORTH) to "yào",
            Syllable("me", Tone.FIFTH) to "me",
            Syllable("zhui", Tone.FIRST) to "zhuī",
            Syllable("liu", Tone.FIRST) to "liū",
            Syllable("nü", Tone.THIRD) to "nǚ",
            Syllable("nu:", Tone.THIRD) to "nǚ",
            Syllable("nv", Tone.THIRD) to "nǚ",
        ) { (syllable, expectedPinyinWithToneMarks) ->
            syllable.format(TransliterationSystem.PINYIN_WITH_TONE_MARKS) shouldBe expectedPinyinWithToneMarks
        }
    }

    context("format to zhuyin with invalid pinyin syllable") {
    }

    context("format to pinyin with tone diacritics with invalid pinyin syllable") {
    }

    context("format to pinyin with tone numbers with invalid pinyin syllable") {
    }
})
