package com.marvinelsen.chinese.transliteration

import io.kotest.assertions.throwables.shouldThrow
import io.kotest.core.spec.style.ShouldSpec
import io.kotest.datatest.withData
import io.kotest.matchers.nulls.shouldBeNull
import io.kotest.matchers.shouldBe

class ToneTest : ShouldSpec({
    context("convert correctly from digit") {
        withData(
            nameFn = { "'${it.first}' -> ${it.second}" },
            '1' to Tone.FIRST,
            '2' to Tone.SECOND,
            '3' to Tone.THIRD,
            '4' to Tone.FORTH,
            '5' to Tone.FIFTH,
        ) { (digit, expectedTone) ->
            Tone.fromDigit(digit) shouldBe expectedTone
        }
    }

    context("convert correctly from Int") {
        withData(
            nameFn = { "'${it.first}' -> ${it.second}" },
            1 to Tone.FIRST,
            2 to Tone.SECOND,
            3 to Tone.THIRD,
            4 to Tone.FORTH,
            5 to Tone.FIFTH,
        ) { (number, expectedTone) ->
            Tone.fromInt(number) shouldBe expectedTone
        }
    }

    context("convert correctly from Zhuyin tone mark") {
        withData(
            nameFn = { "'${it.first}' -> ${it.second}" },
            'ˊ' to Tone.SECOND,
            'ˇ' to Tone.THIRD,
            'ˋ' to Tone.FORTH,
            '˙' to Tone.FIFTH,
        ) { (zhuyinToneMark, expectedTone) ->
            Tone.fromZhuyinToneMark(zhuyinToneMark) shouldBe expectedTone
        }
    }

    context("convert correctly to Int") {
        withData(
            nameFn = { "${it.first} -> '${it.second}'" },
            Tone.FIRST to 1,
            Tone.SECOND to 2,
            Tone.THIRD to 3,
            Tone.FORTH to 4,
            Tone.FIFTH to 5,
        ) { (tone, expectedInteger) ->
            tone.toInt() shouldBe expectedInteger
        }
    }

    context("return null when converting from invalid digit ") {
        withData(
            nameFn = { "'$it' -> null" },
            '0',
            '6',
            'a',
            'z',
            '$',
            '*',
        ) { invalidDigit ->
            Tone.fromDigitOrNull(invalidDigit).shouldBeNull()
        }
    }

    context("return null when converting from invalid int ") {
        withData(
            nameFn = { "'$it' -> throws exception" },
            0,
            6,
            -1,
            Int.MAX_VALUE,
            Int.MIN_VALUE,
        ) { invalidNumber ->
            Tone.fromIntOrNull(invalidNumber).shouldBeNull()
        }
    }

    context("return null when converting from invalid Zhuyin tone mark ") {
        withData(
            nameFn = { "'$it' -> throws exception" },
            '0',
            '6',
            'a',
            'z',
            '$',
            '*',
        ) { invalidZhuyinToneMark ->
            Tone.fromZhuyinToneMarkOrNull(invalidZhuyinToneMark).shouldBeNull()
        }
    }

    context("throw exception when converting from invalid digit ") {
        withData(
            nameFn = { "'$it' -> throws exception" },
            '0',
            '6',
            'a',
            'z',
            '$',
            '*',
        ) { invalidDigit ->
            shouldThrow<IllegalArgumentException> {
                Tone.fromDigit(invalidDigit)
            }
        }
    }

    context("throw exception when converting from invalid int ") {
        withData(
            nameFn = { "'$it' -> throws exception" },
            0,
            6,
            -1,
            Int.MAX_VALUE,
            Int.MIN_VALUE,
        ) { invalidNumber ->
            shouldThrow<IllegalArgumentException> {
                Tone.fromInt(invalidNumber)
            }
        }
    }

    context("throw exception when converting from invalid Zhuyin tone mark ") {
        withData(
            nameFn = { "'$it' -> throws exception" },
            '0',
            '6',
            'a',
            'z',
            '$',
            '*',
        ) { invalidZhuyinToneMark ->
            shouldThrow<IllegalArgumentException> {
                Tone.fromZhuyinToneMark(invalidZhuyinToneMark)
            }
        }
    }

    context("format to Zhuyin correctly") {
        withData(
            nameFn = { "${it.first} -> '${it.second}'" },
            Tone.FIRST to "",
            Tone.SECOND to "ˊ",
            Tone.THIRD to "ˇ",
            Tone.FORTH to "ˋ",
            Tone.FIFTH to "˙",
        ) { (tone, expectedZhuyinToneMark) ->
            tone.format(TransliterationSystem.ZHUYIN) shouldBe expectedZhuyinToneMark
        }
    }

    context("format to Pinyin with tone numbers correctly") {
        withData(
            nameFn = { "${it.first} -> '${it.second}'" },
            Tone.FIRST to "1",
            Tone.SECOND to "2",
            Tone.THIRD to "3",
            Tone.FORTH to "4",
            Tone.FIFTH to "5",
        ) { (tone, expectedNumber) ->
            tone.format(TransliterationSystem.PINYIN_WITH_TONE_NUMBERS) shouldBe expectedNumber
        }
    }

    context("format to Pinyin with tone marks correctly") {
        withData(
            nameFn = { "${it.first} -> '${it.second}'" },
            Tone.FIRST to "\u0304",
            Tone.SECOND to "\u0301",
            Tone.THIRD to "\u030C",
            Tone.FORTH to "\u0300",
            Tone.FIFTH to "",
        ) { (tone, expectedAccent) ->
            tone.format(TransliterationSystem.PINYIN_WITH_TONE_MARKS) shouldBe expectedAccent
        }
    }
})
